@@ -1,5 +1,5 @@
 use std::{
-    convert::{TryFrom, TryInto},
+    convert::TryFrom,
     pin::Pin,
     task::{Context, Poll},
     time::Duration,
@@ -17,8 +17,7 @@ use tracing::debug;
 
 use crate::{
     auth::grpc::{AuthGrpcService, OAuthTokenSource},
-    grpc::StatusCodeSet,
-    pubsub::api,
+    pubsub::{api, PubSubRetryCheck},
     retry_policy::{exponential_backoff, ExponentialBackoff, RetryOperation, RetryPolicy},
 };
 
@@ -40,9 +39,9 @@ config_default! {
         @default(0, "StreamSubscriptionConfig::default_max_outstanding_bytes")
         pub max_outstanding_bytes: i64,
 
-        /// Maximum number of ack ids to buffer while waiting for pull requests to complete. This value
-        /// should be at least as large as the maximum number of outstanding messages.
-        @default(2000, "StreamSubscriptionConfig::default_ack_channel_capacity")
+        /// Deprecated, subsumed by `max_outstanding_messages`
+        #[deprecated]
+        @default(0, "StreamSubscriptionConfig::default_ack_channel_capacity")
         pub ack_channel_capacity: usize,
     }
 }
@@ -257,7 +256,8 @@ fn create_streaming_pull_request_stream(
 /// The stream returned by the
 /// [`stream_subscription`](crate::pubsub::SubscriberClient::stream_subscription) function
 #[pin_project]
-pub struct StreamSubscription<C = crate::DefaultConnector, R = ExponentialBackoff<StatusCodeSet>> {
+pub struct StreamSubscription<C = crate::DefaultConnector, R = ExponentialBackoff<PubSubRetryCheck>>
+{
     state: StreamState<C, R>,
     // reserve the right to be !Unpin in the future without a major version bump
     _p: std::marker::PhantomPinned,
@@ -300,15 +300,26 @@ impl<C> StreamSubscription<C> {
                 subscription,
                 config,
                 retry_policy: ExponentialBackoff::new(
-                    // ensure that the retry codes include Unavailable to retry the pubsub
-                    // documented cases of disconnection
-                    crate::pubsub::DEFAULT_RETRY_CODES | tonic::Code::Unavailable.into(),
-                    exponential_backoff::Config::default(),
+                    PubSubRetryCheck::default(),
+                    Self::default_retry_configuration(),
                 ),
             },
             _p: std::marker::PhantomPinned,
         }
     }
+
+    /// The default configuration values used for retrying connections to the PubSub streaming pull
+    /// RPC
+    pub fn default_retry_configuration() -> exponential_backoff::Config {
+        // values pulled from java lib
+        // https://github.com/googleapis/java-pubsub/blob/d969e8925edc3401e6eb534699ce0351a5f0b20b/google-cloud-pubsub/src/main/java/com/google/cloud/pubsub/v1/StreamingSubscriberConnection.java#L70
+        exponential_backoff::Config {
+            initial_interval: Duration::from_millis(100),
+            max_interval: Duration::from_secs(10),
+            multiplier: 2.0,
+            ..Default::default()
+        }
+    }
 }
 
 impl<C, OldR> StreamSubscription<C, OldR> {
@@ -412,14 +423,6 @@ where
     R::RetryOp: Send + 'static,
     <R::RetryOp as RetryOperation<(), tonic::Status>>::Sleep: Send + 'static,
 {
-    // Verify channel size is valid.
-    assert!(
-        config.ack_channel_capacity
-            >= config
-                .max_outstanding_messages
-                .try_into()
-                .expect("should be a valid usize")
-    );
     let subscription_meta =
         MetadataValue::from_str(&subscription).expect("valid subscription metadata");
 
@@ -432,7 +435,10 @@ where
     // is a period of time during reconnection when two receivers might exist, because the
     // disconnected stream is dropped in a background task at an unknown time (unknown to our
     // layer anyway), and the new receiver should continue to pull from the existing senders.
-    let (sender, receiver) = mpmc::bounded(config.ack_channel_capacity);
+    let (sender, receiver) = mpmc::bounded(
+        usize::try_from(config.max_outstanding_messages)
+            .expect("outstanding messages should fit in usize"),
+    );
 
     async_stream::stream! {
         let mut retry_op = None;
@@ -446,62 +452,77 @@ where
             );
 
             debug!(message="connecting streaming pull stream", %subscription, %client_id);
-            let mut message_stream = client
+            let mut error = match client
                 .streaming_pull(request_stream)
                 .await
-                .map_err(|mut err| {
-                    err.metadata_mut()
-                        .insert("subscription", subscription_meta.clone());
-                    err
-                })?
-                .into_inner();
+                .map(|response| response.into_inner())
+            {
+                Err(err) => err,
+                Ok(mut message_stream) => 'read: loop {
+                    match message_stream.next().await {
+                        // If the stream is out of elements, some connection must have been closed.
+                        // However PubSub docs say StreamingPull always terminates with an error,
+                        // so this normal end-of-stream shouldn't happen, and instead should fall
+                        // to the error branch.
+                        //
+                        // Here we assume some other part of the stack ended the connection, and
+                        // therefore attempting to reconnect (with retry/backoff) is the right
+                        // resolution
+                        None => break 'read tonic::Status::aborted("unexpected end of stream"),
 
-            while let Some(response_result) = message_stream.next().await {
-                match response_result {
-                    Ok(response) => {
-                        // we got a successful response. If we were in a retry loop, declare it
-                        // complete
-                        retry_op = None;
+                        // if there's an error in reading, break to the error handler and check
+                        // whether to retry
+                        Some(Err(err)) => break 'read err,
 
-                        for message in response.received_messages {
-                            let ack_token = AcknowledgeToken {
-                                id: message.ack_id,
-                                channel: sender.clone()
-                            };
-                            let message = message
-                                .message
-                                .expect("message should be populated by RPC server");
-                            yield Ok((ack_token, message));
-                        }
-                    }
-                    Err(mut err) => {
-                        // check if this error can be recovered by reconnecting the stream.
-                        let should_retry = retry_op
-                            .get_or_insert_with(|| retry_policy.new_operation())
-                            .check_retry(&(), &err);
+                        // otherwise, we got a successful response
+                        Some(Ok(response)) => {
+                            // If we were in a retry loop, declare it complete
+                            retry_op = None;
 
-                        match should_retry {
-                            // if the retry policy determines a retry is possible, sleep for the
-                            // given backoff and then try reconnecting
-                            Some(backoff_sleep) => {
-                                backoff_sleep.await;
-                                continue 'reconnect;
-                            }
-                            // if the policy does not provide a sleep, then it determined that the
-                            // operation is terminal (or the retries have been exhausted). Yield
-                            // the error, and then exit
-                            None => {
-                                err.metadata_mut().insert("subscription", subscription_meta);
-                                yield Err(err);
-                                break;
+                            for message in response.received_messages {
+                                let ack_token = AcknowledgeToken {
+                                    id: message.ack_id,
+                                    channel: sender.clone()
+                                };
+                                let message = match message.message {
+                                    Some(msg) => msg,
+                                    None => break 'read tonic::Status::internal(
+                                        "message should be populated by RPC server"
+                                    ),
+                                };
+                                yield Ok((ack_token, message));
                             }
+
+                            continue 'read;
                         }
                     }
                 }
             };
 
-            // the stream is out of elements. exit the reconnect loop
-            break;
+            // if either the streaming connection or a stream element produces an error,
+            // the error will arrive here.
+
+            // check if this error can be recovered by reconnecting the stream.
+            let should_retry = retry_op
+                .get_or_insert_with(|| retry_policy.new_operation())
+                .check_retry(&(), &error);
+
+            match should_retry {
+                // if the retry policy determines a retry is possible, sleep for the
+                // given backoff and then try reconnecting
+                Some(backoff_sleep) => {
+                    backoff_sleep.await;
+                    continue 'reconnect;
+                }
+                // if the policy does not provide a sleep, then it determined that the
+                // operation is terminal (or the retries have been exhausted). Yield
+                // the error, and then exit
+                None => {
+                    error.metadata_mut().insert("subscription", subscription_meta);
+                    yield Err(error);
+                    break 'reconnect;
+                }
+            }
         }
     }
 }
@@ -526,8 +547,8 @@ mod test {
             StreamSubscriptionConfig {
                 max_outstanding_messages: 2000,
                 max_outstanding_bytes: 3000,
-                ack_channel_capacity: 4000,
                 stream_ack_deadline: Duration::from_secs(20),
+                ..Default::default()
             },
         );
         pin_mut!(requests);
